from typing import Any, Callable, Iterable, List, Optional, TYPE_CHECKING, Tuple, Union, Dict, Sequence
import argparse

import matplotlib.pyplot as plt

from sinter import shot_error_rate_to_piece_error_rate
from sinter._main_combine import ExistingData
from sinter._plotting import plot_discard_rate, plot_custom
from sinter._plotting import plot_error_rate

if TYPE_CHECKING:
    import sinter


def parse_args(args: List[str]) -> Any:
    parser = argparse.ArgumentParser(description='Plot collected CSV data.',
                                     prog='sinter plot')
    parser.add_argument('--filter_func',
                        type=str,
                        default="True",
                        help='A python expression that determines whether a case is kept or not.\n'
                             'Values available to the python expression:\n'
                             '    metadata: The parsed value from the json_metadata for the data point.\n'
                             '    decoder: The decoder that decoded the data for the data point.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled for the data point.\n'
                             '    stat: The sinter.TaskStats object for the data point.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `bool` to get True or False.\n'
                             'Examples:\n'
                             '''    --filter_func "decoder=='pymatching'"\n'''
                             '''    --filter_func "0.001 < metadata['p'] < 0.005"\n''')
    parser.add_argument('--x_func',
                        type=str,
                        default="1",
                        help='A python expression that determines where points go on the x axis.\n'
                             'Values available to the python expression:\n'
                             '    metadata: The parsed value from the json_metadata for the data point.\n'
                             '    decoder: The decoder that decoded the data for the data point.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled for the data point.\n'
                             '    stat: The sinter.TaskStats object for the data point.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `float` to get a float.\n'
                             'Examples:\n'
                             '''    --x_func "metadata['p']"\n'''
                             '''    --x_func "metadata['path'].split('/')[-1].split('.')[0]"\n'''
                        )
    parser.add_argument('--y_func',
                        type=str,
                        default=None,
                        help='A python expression that determines where points go on the y axis.\n'
                             'This argument is not used by error rate or discard rate plots; only\n'
                             'by the "custom_y" type plot.'
                             'Values available to the python expression:\n'
                             '    metadata: The parsed value from the json_metadata for the data point.\n'
                             '    decoder: The decoder that decoded the data for the data point.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled for the data point.\n'
                             '    stat: The sinter.TaskStats object for the data point.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `float` to get a float.\n'
                             'Examples:\n'
                             '''    --x_func "metadata['p']"\n'''
                             '''    --x_func "metadata['path'].split('/')[-1].split('.')[0]"\n'''
                        )
    parser.add_argument('--fig_size',
                        type=int,
                        nargs=2,
                        default=None,
                        help='Desired figure width in pixels.')
    parser.add_argument('--fig_height',
                        type=int,
                        default=None,
                        help='Desired figure height in pixels.')
    parser.add_argument('--group_func',
                        type=str,
                        default="'all data (use -group_func and -x_func to group into curves)'",
                        help='A python expression that determines how points are grouped into curves.\n'
                             'Values available to the python expression:\n'
                             '    metadata: The parsed value from the json_metadata for the data point.\n'
                             '    decoder: The decoder that decoded the data for the data point.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled for the data point.\n'
                             '    stat: The sinter.TaskStats object for the data point.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `str` to get a useful string.\n'
                             'Examples:\n'
                             '''    --group_func "(decoder, metadata['d'])"\n'''
                             '''    --group_func "metadata['path'].split('/')[-2]"\n'''
                        )
    parser.add_argument('--failure_unit_name',
                        type=str,
                        default=None,
                        help='The unit of failure, typically either "shot" (the default) or "round".\n'
                             'If this argument is specified, --failure_units_per_shot_func must also be specified.\n'''
                        )
    parser.add_argument('--failure_units_per_shot_func',
                        type=str,
                        default=None,
                        help='A python expression that evaluates to the number of failure units there are per shot.\n'
                             'For example, if the failure unit is rounds, this should be an expression that returns\n'
                             'the number of rounds in a shot. Sinter has no way of knowing what you consider a round\n'
                             'to be, otherwise.\n'
                             '\n'
                             'This value is used to rescale the logical error rate plots. For example, if there are 4\n'
                             'failure units per shot then a shot error rate of 10% corresponds to a unit failure rate\n'
                             'of 2.7129%. The conversion formula (assuming <50% error rates) is:\n'
                             '\n'
                             '    P_unit = 0.5 - 0.5 * (1 - 2 * P_shot)**(1/units_per_shot)\n'
                             '\n'
                             'Values available to the python expression:\n'
                             '    metadata: The parsed value from the json_metadata for the data point.\n'
                             '    decoder: The decoder that decoded the data for the data point.\n'
                             '    strong_id: The cryptographic hash of the case that was sampled for the data point.\n'
                             '    stat: The sinter.TaskStats object for the data point.\n'
                             '\n'
                             'Expected expression type:\n'
                             '    float.\n'
                             '\n'
                             'Examples:\n'
                             '''    --failure_units_per_shot_func "metadata['rounds']"\n'''
                             '''    --failure_units_per_shot_func "metadata['distance'] * 3"\n'''
                             '''    --failure_units_per_shot_func "10"\n'''
                        )
    parser.add_argument('--plot_args_func',
                        type=str,
                        default='''{'marker': 'ov*sp^<>8PhH+xXDd|'[index % 18]}''',
                        help='A python expression used to customize the look of curves.\n'
                             'Values available to the python expression:\n'
                             '    index: A unique integer identifying the curve.\n'
                             '    key: The group key (returned from --group_func) identifying the curve.\n'
                             '    stats: The list of sinter.TaskStats object in the group.\n'
                             '    metadata: (From one arbitrary data point in the group.) The parsed value from the json_metadata for the data point.\n'
                             '    decoder: (From one arbitrary data point in the group.) The decoder that decoded the data for the data point.\n'
                             '    strong_id: (From one arbitrary data point in the group.) The cryptographic hash of the case that was sampled for the data point.\n'
                             '    stat: (From one arbitrary data point in the group.) The sinter.TaskStats object for the data point.\n'
                             'Expected expression type:\n'
                             '    A dictionary to give to matplotlib plotting functions as a **kwargs argument.\n'
                             'Examples:\n'
                             """    --plot_args_func "'''{'label': 'curve #' + str(index), 'linewidth': 5}'''"\n"""
                             """    --plot_args_func "'''{'marker': 'ov*sp^<>8PhH+xXDd|'[index % 18]}'''"\n"""
                        )
    parser.add_argument('--in',
                        type=str,
                        nargs='+',
                        required=True,
                        help='Input files to get data from.')
    parser.add_argument('--type',
                        choices=['error_rate', 'discard_rate', 'custom_y'],
                        nargs='+',
                        default=(),
                        help='Picks the figures to include.')
    parser.add_argument('--out',
                        type=str,
                        default=None,
                        help='Output file to write the plot to.\n'
                             'The file extension determines the type of image.\n'
                             'Either this or --show must be specified.')
    parser.add_argument('--xaxis',
                        type=str,
                        default='[log]',
                        help='Customize the X axis label. Prefix [log] for log scale.')
    parser.add_argument('--yaxis',
                        type=str,
                        default=None,
                        help='Customize the Y axis label. Prefix [log] for log scale.')
    parser.add_argument('--show',
                        action='store_true',
                        help='Displays the plot in a window.\n'
                             'Either this or --out must be specified.')
    parser.add_argument('--ymin',
                        default=None,
                        type=float,
                        help='Sets the minimum value of the y axis (max always 1).')
    parser.add_argument('--title',
                        default=None,
                        type=str,
                        help='Sets the title of the plot.')
    parser.add_argument('--highlight_max_likelihood_factor',
                        type=float,
                        default=1000,
                        help='The relative likelihood ratio that determines the color highlights around curves.\n'
                             'Set this to 1 or larger. Set to 1 to disable highlighting.')

    a = parser.parse_args(args=args)
    if not a.show and a.out is None:
        raise ValueError("Must specify '--out file' or '--show'.")
    if 'custom_y' in a.type and a.y_func is None:
        raise ValueError("--type custom_y requires --y_func.")
    if a.y_func is not None and a.type and 'custom_y' not in a.type:
        raise ValueError("--y_func only works with --type custom_y.")
    if (len(a.type) == 0 and a.y_func is not None) or list(a.type) == 'custom_y':
        if a.failure_units_per_shot_func is not None:
            raise ValueError("--failure_units_per_shot_func doesn't affect --type custom_y")
    if (a.failure_units_per_shot_func is not None) != (a.failure_unit_name is not None):
        raise ValueError("--failure_units_per_shot_func and --failure_unit_name can only be specified together.")
    if a.failure_units_per_shot_func is None:
        a.failure_units_per_shot_func = "1"
    if a.failure_unit_name is None:
        a.failure_unit_name = 'shot'

    a.x_func = eval(compile(
        f'lambda *, stat, decoder, metadata, strong_id: {a.x_func}',
        filename='x_func:command_line_arg',
        mode='eval'))
    if a.y_func is not None:
        a.y_func = eval(compile(
            f'lambda *, stat, decoder, metadata, strong_id: {a.y_func}',
            filename='x_func:command_line_arg',
            mode='eval'))
    a.group_func = eval(compile(
        f'lambda *, stat, decoder, metadata, strong_id: {a.group_func}',
        filename='group_func:command_line_arg',
        mode='eval'))
    a.filter_func = eval(compile(
        f'lambda *, stat, decoder, metadata, strong_id: {a.filter_func}',
        filename='filter_func:command_line_arg',
        mode='eval'))
    a.failure_units_per_shot_func = eval(compile(
        f'lambda *, stat, decoder, metadata, strong_id: {a.failure_units_per_shot_func}',
        filename='failure_units_per_shot_func:command_line_arg',
        mode='eval'))
    a.plot_args_func = eval(compile(
        f'lambda *, index, key, stats, stat, decoder, metadata, strong_id: {a.plot_args_func}',
        filename='plot_args_func:command_line_arg',
        mode='eval'))
    return a


def _ax_log_helper(*, ax: Optional[plt.Axes], log_x: bool, log_y: bool):
    if ax is None:
        return
    if log_x and log_y:
        ax.loglog()
    elif log_y:
        ax.semilogy()
    elif log_x:
        ax.semilogx()


def _plot_helper(
    *,
    samples: Union[Iterable['sinter.TaskStats'], ExistingData],
    group_func: Callable[['sinter.TaskStats'], Any],
    filter_func: Callable[['sinter.TaskStats'], Any],
    failure_units_per_shot_func: Callable[['sinter.TaskStats'], Any],
    x_func: Callable[['sinter.TaskStats'], Any],
    y_func: Optional[Callable[['sinter.TaskStats'], Any]],
    failure_unit: str,
    plot_types: Sequence[str],
    highlight_max_likelihood_factor: Optional[float],
    xaxis: str,
    yaxis: Optional[str],
    min_y: Optional[float],
    title: Optional[str],
    fig_size: Optional[Tuple[int, int]],
    plot_args_func: Callable[[int, Any, List['sinter.TaskStats']], Dict[str, Any]],
) -> Tuple[plt.Figure, List[plt.Axes]]:
    if isinstance(samples, ExistingData):
        total = samples
    else:
        total = ExistingData()
        for sample in samples:
            total.add_sample(sample)
    total.data = {k: v
                  for k, v in total.data.items()
                  if bool(filter_func(v))}

    if not plot_types:
        if y_func is not None:
            plot_types = ['custom_y']
        else:
            plot_types = ['error_rate']
            if any(s.discards for s in total.data.values()):
                plot_types.append('discard_rate')
    include_error_rate_plot = 'error_rate' in plot_types
    include_discard_rate_plot = 'discard_rate' in plot_types
    include_custom_plot = 'custom_y' in plot_types
    num_plots = include_error_rate_plot + include_discard_rate_plot + include_custom_plot

    fig: plt.Figure
    ax_err: Optional[plt.Axes] = None
    ax_dis: Optional[plt.Axes] = None
    ax_cus: Optional[plt.Axes] = None
    fig, axes = plt.subplots(1, num_plots)
    if num_plots == 1:
        axes = [axes]
    axes = list(axes)
    if include_custom_plot:
        ax_cus = axes.pop()
    if include_discard_rate_plot:
        ax_dis = axes.pop()
    if include_error_rate_plot:
        ax_err = axes.pop()

    plotted_stats: List['sinter.TaskStats'] = [
        stat
        for stat in total.data.values()
        if filter_func(stat)
    ]

    want_log_x = xaxis.startswith('[log]')
    force_log_y = yaxis is not None and yaxis.startswith('[log]')
    force_not_log_y = yaxis is not None and not yaxis.startswith('[log]')
    _ax_log_helper(ax=ax_err, log_x=want_log_x, log_y=not force_not_log_y)
    _ax_log_helper(ax=ax_dis, log_x=want_log_x, log_y=force_log_y and ax_err is None)
    _ax_log_helper(ax=ax_cus, log_x=want_log_x, log_y=force_log_y)
    if xaxis.startswith('[log]'):
        xaxis = xaxis[5:]
    if yaxis is not None and yaxis.startswith('[log]'):
        yaxis = yaxis[5:]

    if ax_err is not None:
        plot_error_rate(
            ax=ax_err,
            stats=plotted_stats,
            group_func=group_func,
            x_func=x_func,
            failure_units_per_shot_func=failure_units_per_shot_func,
            highlight_max_likelihood_factor=highlight_max_likelihood_factor,
            plot_args_func=plot_args_func,
        )
        if min_y is None:
            min_y = 1
            for stat in plotted_stats:
                if stat.shots <= stat.discards:
                    continue
                err_rate = stat.errors / (stat.shots - stat.discards)
                pieces = failure_units_per_shot_func(stat)
                err_rate = shot_error_rate_to_piece_error_rate(err_rate, pieces=pieces)
                if err_rate < min_y:
                    min_y = err_rate
            if not plotted_stats:
                min_y = 1e-4
            low_d = 4
            while 10**-low_d > min_y*0.9 and low_d < 10:
                low_d += 1
            min_y = 10**-low_d
        major_tick_steps = 1
        while 10**-major_tick_steps >= min_y * 0.1:
            major_tick_steps += 1
        ax_err.set_yticks([10**-d for d in range(major_tick_steps)])
        ax_err.set_yticks([b*10**-d for d in range(1, major_tick_steps) for b in range(2, 10)], minor=True)
        ax_err.set_ylim(min_y, 1e-0)
        if yaxis is not None and not include_custom_plot:
            ax_err.set_ylabel(yaxis)
        else:
            ax_err.set_ylabel(f"Logical Error Rate (per {failure_unit})")
        ax_err.grid(which='major', color='#000000')
        ax_err.grid(which='minor', color='#DDDDDD')
        ax_err.legend()

    if ax_dis is not None:
        plot_discard_rate(
            ax=ax_dis,
            stats=plotted_stats,
            group_func=group_func,
            failure_units_per_shot_func=failure_units_per_shot_func,
            x_func=x_func,
            highlight_max_likelihood_factor=highlight_max_likelihood_factor,
            plot_args_func=plot_args_func,
        )
        ax_dis.set_yticks([p / 10 for p in range(11)], labels=[f'{10*p}%' for p in range(11)])
        ax_dis.set_ylim(0, 1)
        ax_dis.grid(which='minor')
        ax_dis.grid(which='major', color='black')
        if yaxis is not None and not include_custom_plot and ax_err is None:
            ax_dis.set_ylabel(yaxis)
        else:
            ax_dis.set_ylabel(f"Discard Rate (per {failure_unit})")
        ax_dis.legend()

    if ax_cus is not None:
        assert y_func is not None
        plot_custom(
            ax=ax_cus,
            stats=plotted_stats,
            x_func=x_func,
            y_func=y_func,
            group_func=group_func,
            filter_func=filter_func,
            plot_args_func=plot_args_func,
        )
        ax_cus.grid()
        if yaxis is not None:
            ax_cus.set_ylabel(yaxis)
        else:
            ax_cus.set_ylabel('custom')
        ax_cus.legend()

    inferred_x_axis = xaxis if xaxis is not None else 'custom'
    for ax in [ax_err, ax_dis, ax_cus]:
        if ax is not None:
            ax.set_xlabel(inferred_x_axis)

    vs_suffix = ''
    if xaxis is not None:
        vs_suffix = f' vs {xaxis}'
    if ax_err is not None:
        ax_err.set_title(f'Logical Error Rate per {failure_unit}{vs_suffix}')
        if title is not None:
            ax_err.set_title(title)
    if ax_dis is not None:
        ax_dis.set_title(f'Discard Rate per {failure_unit}{vs_suffix}')
    if ax_cus is not None:
        if title is not None:
            ax_cus.set_title(title)
        else:
            ax_cus.set_title(f'Custom Plot')

    if fig_size is None:
        fig.set_size_inches(10 * num_plots, 10)
        fig.set_dpi(100)
    else:
        w, h = fig_size
        fig.set_size_inches(w / 100, h / 100)
        fig.set_dpi(100)
    fig.tight_layout()
    axs = [e for e in [ax_err, ax_dis] if e is not None]
    return fig, axs


def main_plot(*, command_line_args: List[str]):
    args = parse_args(command_line_args)
    total = ExistingData()
    for file in getattr(args, 'in'):
        total += ExistingData.from_file(file)

    fig, _ = _plot_helper(
        samples=total,
        group_func=lambda stat: args.group_func(
            stat=stat,
            decoder=stat.decoder,
            metadata=stat.json_metadata,
            strong_id=stat.strong_id),
        x_func=lambda stat: args.x_func(
            stat=stat,
            decoder=stat.decoder,
            metadata=stat.json_metadata,
            strong_id=stat.strong_id),
        y_func=None if args.y_func is None else lambda stat: args.y_func(
            stat=stat,
            decoder=stat.decoder,
            metadata=stat.json_metadata,
            strong_id=stat.strong_id),
        filter_func=lambda stat: args.filter_func(
            stat=stat,
            decoder=stat.decoder,
            metadata=stat.json_metadata,
            strong_id=stat.strong_id),
        failure_units_per_shot_func=lambda stat: args.failure_units_per_shot_func(
            stat=stat,
            decoder=stat.decoder,
            metadata=stat.json_metadata,
            strong_id=stat.strong_id),
        plot_args_func=lambda index, group_key, stats: args.plot_args_func(
            index=index,
            key=group_key,
            stats=stats,
            stat=stats[0],
            decoder=stats[0].decoder,
            metadata=stats[0].json_metadata,
            strong_id=stats[0].strong_id),
        failure_unit=args.failure_unit_name,
        plot_types=args.type,
        xaxis=args.xaxis,
        yaxis=args.yaxis,
        fig_size=args.fig_size,
        min_y=args.ymin,
        highlight_max_likelihood_factor=args.highlight_max_likelihood_factor,
        title=args.title,
    )
    if args.out is not None:
        fig.savefig(args.out)
    if args.show:
        plt.show()



import contextlib
import io
import pathlib
import tempfile

import pytest
from sinter._main import main


def test_main_plot():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "int('a' in metadata['path'])",
                "--group_func",
                "decoder",
                "--ymin",
                "1e-3",
                "--title",
                "test_plot",
            ])
        assert (d / "output.png").exists()


def test_main_plot_2():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""path"":""a.stim""}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""path"":""b.stim""}"
            """.strip(), file=f)

        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "int('a' in metadata['path'])",
                "--group_func",
                "decoder",
                "--plot_args_func",
                "{'lw': 2, 'color': 'r' if decoder == 'never' else 'b'}",
            ])
        assert (d / "output.png").exists()


def test_main_plot_failure_units():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""r"":15,""d"":5}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf7,"{""r"":9,""d"":3}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""r"":6,""d"":2}"
            """.strip(), file=f)

        out = io.StringIO()
        with pytest.raises(ValueError,  match="specified together"):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "metadata['d']",
                "--group_func",
                "decoder",
                "--failure_units_per_shot_func",
                "metadata['r']",
            ])
        with pytest.raises(ValueError,  match="specified together"):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "metadata['d']",
                "--group_func",
                "decoder",
                "--failure_unit_name",
                "Rounds",
            ])
        assert not (d / "output.png").exists()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "metadata['d']",
                "--group_func",
                "decoder",
                "--failure_units_per_shot_func",
                "metadata['r']",
                "--failure_unit_name",
                "Rounds",
            ])
        assert (d / "output.png").exists()


def test_main_plot_custom_y_func():
    with tempfile.TemporaryDirectory() as d:
        d = pathlib.Path(d)
        with open(d / f'input.csv', 'w') as f:
            print("""
shots,errors,discards,seconds,decoder,strong_id,json_metadata
300,1,20,1.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf6,"{""r"":15,""d"":5}"
300,100,200,2.0,pymatching,f256bab362f516ebe4d59a08ae67330ff7771ff738757cd738f4b30605ddccf7,"{""r"":9,""d"":3}"
9,5,4,6.0,pymatching,5fe5a6cd4226b1a910d57e5479d1ba6572e0b3115983c9516360916d1670000f,"{""r"":6,""d"":2}"
            """.strip(), file=f)

        out = io.StringIO()
        with pytest.raises(AttributeError,  match="secondsX"):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "metadata['d']",
                "--group_func",
                "decoder",
                "--y_func",
                "stat.secondsX",
                "--yaxis",
                "test axis"
            ])
        assert not (d / "output.png").exists()
        with contextlib.redirect_stdout(out):
            main(command_line_args=[
                "plot",
                "--in",
                str(d / "input.csv"),
                "--out",
                str(d / "output.png"),
                "--x_func",
                "metadata['d']",
                "--group_func",
                "decoder",
                "--y_func",
                "stat.seconds",
                "--yaxis",
                "test axis"
            ])
        assert (d / "output.png").exists()



import argparse
import math
import sys
from typing import Iterator, Any, Tuple, List, Callable, FrozenSet, Optional
from typing import cast

import numpy as np
import stim

import sinter
from sinter._printer import ThrottledProgressPrinter
from sinter._task import Task
from sinter._collection import collect, Progress, post_selection_mask_from_predicate
from sinter._decoding_all_built_in_decoders import BUILT_IN_DECODERS
from sinter._main_combine import ExistingData, CSV_HEADER


def iter_file_paths_into_goals(circuit_paths: Iterator[str],
                               metadata_func: Callable,
                               postselected_detectors_predicate: Optional[Callable[[int, Any, Tuple[float, ...]], bool]],
                               postselected_observables_predicate: Callable[[int, Any], bool],
                               ) -> Iterator[Task]:
    for path in circuit_paths:
        with open(path) as f:
            circuit_text = f.read()
        circuit = stim.Circuit(circuit_text)

        metadata = metadata_func(path=path, circuit=circuit)
        if postselected_detectors_predicate is not None:
            post_mask = post_selection_mask_from_predicate(circuit, metadata=metadata, postselected_detectors_predicate=postselected_detectors_predicate)
            if not np.any(post_mask):
                post_mask = None
        else:
            post_mask = None
        postselected_observables = [
            k
            for k in range(circuit.num_observables)
            if postselected_observables_predicate(k, metadata)
        ]
        if any(postselected_observables):
            postselected_observables_mask = np.zeros(shape=math.ceil(circuit.num_observables / 8), dtype=np.uint8)
            for k in postselected_observables:
                postselected_observables_mask[k // 8] |= 1 << (k % 8)
        else:
            postselected_observables_mask = None

        yield Task(
            circuit=circuit,
            postselection_mask=post_mask,
            postselected_observables_mask=postselected_observables_mask,
            json_metadata=metadata,
        )


def parse_args(args: List[str]) -> Any:
    parser = argparse.ArgumentParser(description='Collect Monte Carlo samples.',
                                     prog='sinter collect')
    parser.add_argument('--circuits',
                        nargs='+',
                        required=True,
                        help='Circuit files to sample from and decode.\n'
                             'This parameter can be given multiple arguments.')
    parser.add_argument('--decoders',
                        type=str,
                        nargs='+',
                        required=True,
                        help='The decoder to use to predict observables from detection events.')
    parser.add_argument('--custom_decoders_module_function',
                        default=None,
                        help='Use the syntax "module:function" to "import function from module" '
                             'and use the result of "function()" as the custom_decoders '
                             'dictionary. The dictionary must map strings to stim.Decoder '
                             'instances.')
    parser.add_argument('--max_shots',
                        type=int,
                        default=None,
                        help='Sampling of a circuit will stop if this many shots have been taken.')
    parser.add_argument('--max_errors',
                        type=int,
                        default=None,
                        help='Sampling of a circuit will stop if this many errors have been seen.')
    parser.add_argument('--processes',
                        required=True,
                        type=int,
                        help='Number of processes to use for simultaneous sampling and decoding.')
    parser.add_argument('--save_resume_filepath',
                        type=str,
                        default=None,
                        help='Activates MERGE mode.\n'
                             "If save_resume_filepath doesn't exist, initializes it with a CSV header.\n"
                             'CSV data already at save_resume_filepath counts towards max_shots and max_errors.\n'
                             'Collected data is appended to save_resume_filepath.\n'
                             'Note that MERGE mode is tolerant to failures: if the process is killed, it can simply be restarted and it will pick up where it left off.\n'
                             'Note that MERGE mode is idempotent: if sufficient data has been collected, no additional work is done when run again.')

    parser.add_argument('--start_batch_size',
                        type=int,
                        default=100,
                        help='Initial number of samples to batch together into one job.\n'
                             'Starting small prevents over-sampling of circuits above threshold.\n'
                             'The allowed batch size increases exponentially from this starting point.')
    parser.add_argument('--max_batch_size',
                        type=int,
                        default=None,
                        help='Maximum number of samples to batch together into one job.\n'
                             'Bigger values increase the delay between jobs finishing.\n'
                             'Smaller values decrease the amount of aggregation of results, increasing the amount of output information.')
    parser.add_argument('--max_batch_seconds',
                        type=int,
                        default=None,
                        help='Limits number of shots in a batch so that the estimated runtime of the batch is below this amount.')
    parser.add_argument('--postselect_detectors_with_non_zero_4th_coord',
                        help='Turns on detector postselection. '
                             'If any detector with a non-zero 4th coordinate fires, the shot is discarded.',
                        action='store_true')
    parser.add_argument('--postselected_detectors_predicate',
                        type=str,
                        default='''False''',
                        help='Specifies a predicate used to decide which detectors to postselect. '
                             'When a postselected detector produces a detection event, the shot is discarded instead of being given to the decoder.'
                             'The number of discarded shots is tracked as a statistic.'
                             'Available values:\n'
                             '    index: The unique number identifying the detector, determined by the order of detectors in the circuit file.\n'
                             '    coords: The coordinate data associated with the detector. An empty tuple, if the circuit file did not specify detector coordinates.\n'
                             '    metadata: The metadata associated with the task being sampled.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `bool` to get False (do not postselect) or True (yes postselect).\n'
                             'Examples:\n'
                             '''    --postselected_detectors_predicate "coords[2] == 0"\n'''
                             '''    --postselected_observables_predicate "coords[3] < metadata['postselection_level']"\n''')
    parser.add_argument('--postselected_observables_predicate',
                        type=str,
                        default='''False''',
                        help='Specifies a predicate used to decide which observables to postselect. '
                             'When a decoder mispredicts a postselected observable, the shot is discarded instead of counting as an error.'
                             'Available values:\n'
                             '    index: The index of the observable to postselect or not.\n'
                             '    metadata: The metadata associated with the task.\n'
                             'Expected expression type:\n'
                             '    Something that can be given to `bool` to get False (do not postselect) or True (yes postselect).\n'
                             'Examples:\n'
                             '''    --postselected_observables_predicate "False"\n'''
                             '''    --postselected_observables_predicate "metadata['d'] == 5 and index >= 2"\n''')
    parser.add_argument('--quiet',
                        help='Disables writing progress to stderr.',
                        action='store_true')
    parser.add_argument('--also_print_results_to_stdout',
                        help='Even if writing to a file, also write results to stdout.',
                        action='store_true')
    parser.add_argument('--existing_data_filepaths',
                        nargs='*',
                        type=str,
                        default=(),
                        help='CSV data from these files counts towards max_shots and max_errors.\n'
                             'This parameter can be given multiple arguments.')
    parser.add_argument('--metadata_func',
                        type=str,
                        default="{'path': path}",
                        help='A python expression that associates json metadata with a circuit\'s results.\n'
                             'Set to "auto" to use "sinter.comma_separated_key_values(path)"\n'
                             'Values available to the expression:\n'
                             '    path: Relative path to the circuit file, from the command line arguments.\n'
                             '    circuit: The circuit itself, parsed from the file, as a stim.Circuit.\n'
                             'Expected type:\n'
                             '    A value that can be serialized into JSON, like a Dict[str, int].\n'
                             '\n'
                             '    Note that the decoder field is already recorded separately, so storing\n'
                             '    it in the metadata as well would be redundant. But something like\n'
                             '    decoder version could be usefully added.\n'
                             'Examples:\n'
                             '''    --metadata_func "{'path': path}"\n'''
                             '''    --metadata_func "auto"\n'''
                             '''    --metadata_func "{'n': circuit.num_qubits, 'p': float(path.split('/')[-1].split('.')[0])}"\n'''
                        )
    a = parser.parse_args(args=args)
    if a.metadata_func == 'auto':
        a.metadata_func = "sinter.comma_separated_key_values(path)"
    a.metadata_func = eval(compile(
        'lambda *, path, circuit: ' + a.metadata_func,
        filename='metadata_func:command_line_arg',
        mode='eval'), {'sinter': sinter})
    a.postselected_observables_predicate = eval(compile(
        'lambda index, metadata: ' + a.postselected_observables_predicate,
        filename='postselected_observables_predicate:command_line_arg',
        mode='eval'))
    if a.postselected_detectors_predicate == 'False':
        if a.postselect_detectors_with_non_zero_4th_coord:
            a.postselected_detectors_predicate = lambda index, metadata, coords: coords[3]
        else:
            a.postselected_detectors_predicate = None
    else:
        if a.postselect_detectors_with_non_zero_4th_coord:
            raise ValueError("Can't specify both --postselect_detectors_with_non_zero_4th_coord and --postselected_detectors_predicate")
        a.postselected_detectors_predicate = eval(compile(
            'lambda index, metadata, coords: ' + cast(str, a.postselected_detectors_predicate),
            filename='postselected_detectors_predicate:command_line_arg',
            mode='eval'))
    if a.custom_decoders_module_function is not None:
        terms = a.custom_decoders_module_function.split(':')
        if len(terms) != 2:
            raise ValueError("--custom_decoders_module_function didn't have exactly one colon "
                             "separating a module name from a function name. Expected an argument "
                             "of the form --custom_decoders_module_function 'module:function'")
        module, function = terms
        vals = {'__name__': '[]'}
        exec(f"from {module} import {function} as _custom_decoders", vals)
        a.custom_decoders = vals['_custom_decoders']()
    else:
        a.custom_decoders = None
    for decoder in a.decoders:
        if decoder not in BUILT_IN_DECODERS and (a.custom_decoders is None or decoder not in a.custom_decoders):
            message = f"Not a recognized decoder: {decoder=}.\n"
            message += f"Available built-in decoders: {sorted(e for e in BUILT_IN_DECODERS.keys() if 'internal' not in e)}.\n"
            if a.custom_decoders is None:
                message += f"No custom decoders are available. --custom_decoders_module_function wasn't specified."
            else:
                message += f"Available custom decoders: {sorted(a.custom_decoders.keys())}."
            raise ValueError(message)

    return a


def open_merge_file(path: str) -> Tuple[Any, ExistingData]:
    try:
        existing = ExistingData.from_file(path)
        return open(path, 'a'), existing
    except FileNotFoundError:
        f = open(path, 'w')
        print(CSV_HEADER, file=f)
        return f, ExistingData()


def main_collect(*, command_line_args: List[str]):
    args = parse_args(args=command_line_args)

    iter_tasks = iter_file_paths_into_goals(
        circuit_paths=args.circuits,
        metadata_func=args.metadata_func,
        postselected_detectors_predicate=args.postselected_detectors_predicate,
        postselected_observables_predicate=args.postselected_observables_predicate,
    )
    num_tasks = len(args.circuits) * len(args.decoders)

    print_to_stdout = args.also_print_results_to_stdout or args.save_resume_filepath is None

    did_work = False
    printer = ThrottledProgressPrinter(
        outs=[],
        print_progress=not args.quiet,
        min_progress_delay=0.03 if args.also_print_results_to_stdout else 0.2,
    )
    if print_to_stdout:
        printer.outs.append(sys.stdout)

    def on_progress(sample: Progress) -> None:
        nonlocal did_work
        for stats in sample.new_stats:
            if not did_work:
                printer.print_out(CSV_HEADER)
                did_work = True
            printer.print_out(stats.to_csv_line())

        msg = sample.status_message
        if msg == 'KeyboardInterrupt':
            printer.show_latest_progress('\nInterrupted. Output is flushed. Cleaning up workers...')
            printer.flush()
        else:
            printer.show_latest_progress(msg)

    try:
        collect(
            num_workers=args.processes,
            hint_num_tasks=num_tasks,
            tasks=iter_tasks,
            print_progress=False,
            save_resume_filepath=args.save_resume_filepath,
            existing_data_filepaths=args.existing_data_filepaths,
            progress_callback=on_progress,
            max_errors=args.max_errors,
            max_shots=args.max_shots,
            decoders=args.decoders,
            max_batch_seconds=args.max_batch_seconds,
            max_batch_size=args.max_batch_size,
            start_batch_size=args.start_batch_size,
            custom_decoders=args.custom_decoders,
        )
    except KeyboardInterrupt:
        pass



import math
import pathlib
from typing import Callable, List, TYPE_CHECKING

import numpy as np
import stim

if TYPE_CHECKING:
    import networkx as nx
    import pymatching


def decode_using_pymatching_v1(*,
                               num_shots: int,
                               num_dets: int,
                               num_obs: int,
                               dem_path: pathlib.Path,
                               dets_b8_in_path: pathlib.Path,
                               obs_predictions_b8_out_path: pathlib.Path,
                               tmp_dir: pathlib.Path,
                               ) -> None:
    """Use pymatching to predict observables from detection events."""

    error_model = stim.DetectorErrorModel.from_file(dem_path)
    matching_graph = detector_error_model_to_pymatching_graph(error_model)
    num_det_bytes = math.ceil(num_dets / 8)

    # note: extra 2 are the boundary node and the invincible-observable-boundary-edge node
    det_bits_buffer = np.zeros(num_dets + 2, dtype=np.bool8)
    with open(dets_b8_in_path, 'rb') as dets_in_f:
        with open(obs_predictions_b8_out_path, 'wb') as obs_out_f:
            for _ in range(num_shots):
                det_bytes = np.fromfile(dets_in_f, dtype=np.uint8, count=num_det_bytes)
                det_bits_buffer[:num_dets] = np.unpackbits(det_bytes, count=num_dets, bitorder='little')
                prediction_bits = matching_graph.decode(det_bits_buffer)
                np.packbits(prediction_bits, bitorder='little').tofile(obs_out_f)


def iter_flatten_model(model: stim.DetectorErrorModel,
                       handle_error: Callable[[float, List[int], List[int]], None],
                       handle_detector_coords: Callable[[int, np.ndarray], None]):
    det_offset = 0
    coords_offset = np.zeros(100, dtype=np.float64)

    def _helper(m: stim.DetectorErrorModel, reps: int):
        nonlocal det_offset
        nonlocal coords_offset
        for _ in range(reps):
            for instruction in m:
                if isinstance(instruction, stim.DemRepeatBlock):
                    _helper(instruction.body_copy(), instruction.repeat_count)
                elif isinstance(instruction, stim.DemInstruction):
                    if instruction.type == "error":
                        dets: List[int] = []
                        frames: List[int] = []
                        t: stim.DemTarget
                        p = instruction.args_copy()[0]
                        for t in instruction.targets_copy():
                            if t.is_relative_detector_id():
                                dets.append(t.val + det_offset)
                            elif t.is_logical_observable_id():
                                frames.append(t.val)
                            elif t.is_separator():
                                # Treat each component of a decomposed error as an independent error.
                                # (Ideally we could configure some sort of correlated analysis; oh well.)
                                handle_error(p, dets, frames)
                                frames = []
                                dets = []
                        # Handle last component.
                        handle_error(p, dets, frames)
                    elif instruction.type == "shift_detectors":
                        det_offset += instruction.targets_copy()[0]
                        a = np.array(instruction.args_copy())
                        coords_offset[:len(a)] += a
                    elif instruction.type == "detector":
                        a = np.array(instruction.args_copy())
                        for t in instruction.targets_copy():
                            handle_detector_coords(t.val + det_offset, a + coords_offset[:len(a)])
                    elif instruction.type == "logical_observable":
                        pass
                    else:
                        raise NotImplementedError()
                else:
                    raise NotImplementedError()
    _helper(model, 1)


def detector_error_model_to_nx_graph(model: stim.DetectorErrorModel) -> 'nx.Graph':
    """Convert a stim error model into a NetworkX graph."""

    # Local import to reduce sinter's startup time.
    try:
        import networkx as nx
    except ImportError as ex:
        raise ImportError(
            "pymatching was installed without networkx?"
            "Run `pip install networkx`.\n"
        ) from ex

    g = nx.Graph()
    boundary_node = model.num_detectors
    g.add_node(boundary_node, is_boundary=True, coords=[-1, -1, -1])

    def handle_error(p: float, dets: List[int], frame_changes: List[int]):
        if p == 0:
            return
        if len(dets) == 0:
            # No symptoms for this error.
            # Code probably has distance 1.
            # Accept it and keep going, though of course decoding will probably perform terribly.
            return
        if len(dets) == 1:
            dets = [dets[0], boundary_node]
        if len(dets) > 2:
            raise NotImplementedError(
                f"Error with more than 2 symptoms can't become an edge or boundary edge: {dets!r}.")
        if g.has_edge(*dets):
            edge_data = g.get_edge_data(*dets)
            old_p = edge_data["error_probability"]
            old_frame_changes = edge_data["fault_ids"]
            # If frame changes differ, the code has distance 2; just keep whichever was first.
            if set(old_frame_changes) == set(frame_changes):
                p = p * (1 - old_p) + old_p * (1 - p)
                g.remove_edge(*dets)
        if p > 0.5:
            p = 1 - p
        if p > 0:
            g.add_edge(*dets, weight=math.log((1 - p) / p), fault_ids=frame_changes, error_probability=p)

    def handle_detector_coords(detector: int, coords: np.ndarray):
        g.add_node(detector, coords=coords)

    iter_flatten_model(model, handle_error=handle_error, handle_detector_coords=handle_detector_coords)

    return g


def detector_error_model_to_pymatching_graph(model: stim.DetectorErrorModel) -> 'pymatching.Matching':
    """Convert a stim error model into a pymatching graph."""

    # Local import to reduce sinter's startup time.
    import pymatching

    g = detector_error_model_to_nx_graph(model)
    num_detectors = model.num_detectors
    num_observables = model.num_observables

    # Ensure invincible detectors are seen by explicitly adding a node for each detector.
    for k in range(num_detectors):
        g.add_node(k)
    # Ensure invincible observables are seen by adding a boundary edge with all observables.
    g.add_node(num_detectors + 1)
    g.add_edge(num_detectors, num_detectors + 1, weight=1, fault_ids=list(range(num_observables)))

    return pymatching.Matching(g)



import os.path
import pathlib

import math
import numpy as np
import stim
import tempfile
from typing import Optional, Union, Dict, TYPE_CHECKING

from sinter._collection import post_selection_mask_from_4th_coord
from sinter._decoding_decoder_class import Decoder
from sinter._decoding_all_built_in_decoders import BUILT_IN_DECODERS
from sinter._decoding import streaming_post_select

if TYPE_CHECKING:
    import sinter


def _converted_on_disk(
        in_path: pathlib.Path,
        out_path: pathlib.Path,
        num_dets: int,
        num_obs: int,
        in_format: str,
        out_format: str) -> pathlib.Path:
    if in_format == out_format:
        return in_path
    raw = stim.read_shot_data_file(
        path=str(in_path),
        format=in_format,
        bit_pack=True,
        num_detectors=num_dets,
        num_observables=num_obs,
    )
    stim.write_shot_data_file(
        data=raw,
        path=str(out_path),
        format=out_format,
        num_detectors=num_dets,
        num_observables=num_obs,
    )
    return out_path


def predict_on_disk(
    *,
    decoder: str,
    dem_path: Union[str, pathlib.Path],
    dets_path: Union[str, pathlib.Path],
    dets_format: str,
    obs_out_path: Union[str, pathlib.Path],
    obs_out_format: str,
    postselect_detectors_with_non_zero_4th_coord: bool = False,
    discards_out_path: Optional[Union[str, pathlib.Path]] = None,
    discards_out_format: Optional[str] = None,
    custom_decoders: Dict[str, 'sinter.Decoder'] = None,
) -> None:
    """Performs decoding and postselection on disk.

    Args:
        decoder: The decoder to use for decoding.
        dem_path: The detector error model to use to configure the decoder.
        dets_path: Where the detection event data is stored on disk.
        dets_format: The format the detection event data is stored in (e.g. '01' or 'b8').
        obs_out_path: Where to write predicted observable flip data on disk.
            Note that the predicted observable flip data will not included data from shots discarded by postselection.
            Use the data in discards_out_path to determine which shots were discarded.
        obs_out_format: The format to write the observable flip data in (e.g. '01' or 'b8').
        postselect_detectors_with_non_zero_4th_coord: Activates postselection. Detectors that have a non-zero 4th
            coordinate will be postselected. Any shot where a postselected detector fires will be discarded.
            Requires specifying discards_out_path, for indicating which shots were discarded.
        discards_out_path: Only used if postselection is being used. Where to write discard data on disk.
        discards_out_format: The format to write discard data in (e.g. '01' or 'b8').
        custom_decoders: Custom decoders that can be used if requested by name.
    """
    if (discards_out_path is not None) != (discards_out_format is not None):
        raise ValueError('(discards_out_path is not None) != (discards_out_format is not None)')
    if (discards_out_path is not None) != postselect_detectors_with_non_zero_4th_coord:
        raise ValueError('(discards_out_path is not None) != postselect_detectors_with_non_zero_4th_coord')

    dem_path = pathlib.Path(dem_path)
    dets_path = pathlib.Path(dets_path)
    obs_out_path = pathlib.Path(obs_out_path)
    if discards_out_path is not None:
        discards_out_path = pathlib.Path(discards_out_path)

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        decode_obj: Optional[Decoder] = None
        if custom_decoders is not None:
            decode_obj = custom_decoders.get(decoder)
        if decode_obj is None:
            decode_obj = BUILT_IN_DECODERS.get(decoder)
        if decode_obj is None:
            raise NotImplementedError(f"Unrecognized decoder: {decoder!r}")

        with open(dem_path) as f:
            dem = stim.DetectorErrorModel(f.read())

        num_dets = dem.num_detectors
        num_det_bytes = math.ceil(num_dets / 8)
        num_obs = dem.num_observables

        dets_b8_path = _converted_on_disk(
            in_path=dets_path,
            out_path=tmp_dir / 'sinter_dets.b8',
            in_format=dets_format,
            out_format='b8',
            num_dets=num_dets,
            num_obs=0)
        if num_det_bytes == 0:
            raise NotImplementedError("Don't know how many shots there are, because num_det_bytes=0.")
        num_shots = os.path.getsize(dets_b8_path) // num_det_bytes

        if discards_out_path is not None:
            if discards_out_format == 'b8':
                discards_b8_path = discards_out_path
            else:
                discards_b8_path = tmp_dir / 'sinter_discards.b8'
            post_selection_mask = np.zeros(dtype=np.uint8, shape=math.ceil(num_dets / 8))
            if postselect_detectors_with_non_zero_4th_coord:
                post_selection_mask = post_selection_mask_from_4th_coord(dem)
            kept_dets_b8_path = tmp_dir / 'sinter_dets.kept.b8'
            num_discards = streaming_post_select(
                num_shots=num_shots,
                num_dets=num_dets,
                num_obs=num_obs,
                dets_in_b8=dets_b8_path,
                obs_in_b8=None,
                obs_out_b8=None,
                discards_out_b8=discards_b8_path,
                dets_out_b8=kept_dets_b8_path,
                post_mask=post_selection_mask,
            )
            assert discards_out_format is not None
            _converted_on_disk(
                in_path=discards_b8_path,
                out_path=discards_out_path,
                out_format=discards_out_format,
                in_format='b8',
                num_dets=1,
                num_obs=0,
            )
            num_kept_shots = num_shots - num_discards
        else:
            kept_dets_b8_path = dets_b8_path
            num_kept_shots = num_shots
            if postselect_detectors_with_non_zero_4th_coord:
                raise ValueError('postselect_detectors_with_non_zero_4th_coord and discards_out_path is None')

        if obs_out_format != 'b8':
            obs_inter = tmp_dir / 'sinter_obs_inter.b8'
        else:
            obs_inter = obs_out_path
        decode_obj.decode_via_files(
            num_shots=num_kept_shots,
            num_dets=num_dets,
            num_obs=num_obs,
            dem_path=dem_path,
            dets_b8_in_path=kept_dets_b8_path,
            obs_predictions_b8_out_path=obs_inter,
            tmp_dir=tmp_dir,
        )
        _converted_on_disk(
            in_path=obs_inter,
            out_path=obs_out_path,
            out_format=obs_out_format,
            in_format='b8',
            num_dets=0,
            num_obs=num_obs,
        )


def predict_discards_bit_packed(
    *,
    dem: stim.DetectorErrorModel,
    dets_bit_packed: np.ndarray,
    postselect_detectors_with_non_zero_4th_coord: bool,
) -> np.ndarray:
    """Determines which shots to discard due to postselected detectors firing.

    Args:
        dem: The detector error model the detector data applies to.
            This is also where coordinate data is read from, in order to determine
            which detectors to postselect as not having fired.
        dets_bit_packed: A uint8 numpy array with shape (num_shots, math.ceil(num_dets / 8)).
            Contains bit packed detection event data.
        postselect_detectors_with_non_zero_4th_coord: Determines how postselection is done.
            Currently, this is the only option so it has to be set to True.
            Any detector from the detector error model that specifies coordinate data with
            at least four coordinates where the fourth coordinate (coord index 3) is non-zero
            will be postselected.

    Returns:
        A numpy bool8 array with shape (num_shots,) where False means not discarded and
        True means yes discarded.
    """
    if not postselect_detectors_with_non_zero_4th_coord:
        raise ValueError("not postselect_detectors_with_non_zero_4th_coord")
    num_dets = dem.num_detectors
    nb = math.ceil(num_dets / 8)
    if len(dets_bit_packed.shape) != 2:
        raise ValueError(f'len(dets_data_bit_packed.shape={dets_bit_packed.shape}) != 2')
    if dets_bit_packed.shape[1] != nb:
        raise ValueError(f'dets_data_bit_packed.shape[1]={dets_bit_packed.shape[1]} != math.ceil(dem.num_detectors={dem.num_detectors} / 8)')
    if dets_bit_packed.dtype != np.uint8:
        raise ValueError(f'dets_data_bit_packed.dtype={dets_bit_packed.dtype} != np.uint8')

    post_selection_mask = np.zeros(dtype=np.uint8, shape=nb)
    if postselect_detectors_with_non_zero_4th_coord:
        for k, coord in dem.get_detector_coordinates().items():
            if len(coord) >= 4 and coord[3]:
                post_selection_mask[k // 8] |= 1 << (k % 8)
    return np.any(dets_bit_packed & post_selection_mask, axis=1)


def predict_observables(
    *,
    dem: stim.DetectorErrorModel,
    dets: np.ndarray,
    decoder: str,
    bit_pack_result: bool = False,
) -> np.ndarray:
    """Predicts which observables were flipped based on detection event data by using a decoder.

    Args:
        dem: The detector error model the detector data applies to.
            This is also where coordinate data is read from, in order to determine
            which detectors to postselect as not having fired.
        dets: The detection event data. Can be bit packed or not bit packed.
            If dtype=np.bool8 then shape=(num_shots, num_detectors).
            If dtype=np.uint8 then shape=(num_shots, math.ceil(num_detectors / 8)).
        decoder: The decoder to use for decoding, e.g. "pymatching".
        bit_pack_result: Defaults to False. Determines if the result is bit packed
            or not.

    Returns:
        If bit_packed_result=False (default):
            dtype=np.bool8
            shape=(num_shots, num_observables)
        If bit_packed_result=True:
            dtype=np.uint8
            shape=(num_shots, math.ceil(num_observables / 8))
    """

    if dets.dtype == np.bool8:
        dets = np.packbits(dets, axis=1, bitorder='little')
    result = predict_observables_bit_packed(dem=dem, dets_bit_packed=dets, decoder=decoder)
    if not bit_pack_result:
        result = np.unpackbits(result, axis=1, bitorder='little', count=dem.num_observables)
    return result


def predict_observables_bit_packed(
    *,
    dem: stim.DetectorErrorModel,
    dets_bit_packed: np.ndarray,
    decoder: str,
    custom_decoders: Optional[Dict[str, 'sinter.Decoder']] = None,
) -> np.ndarray:
    """Predicts which observables were flipped based on detection event data by using a decoder.

    This is a specialization of `sinter.predict_observables`, without optional bit packing.

    Args:
        dem: The detector error model the detector data applies to.
            This is also where coordinate data is read from, in order to determine
            which detectors to postselect as not having fired.
        dets_bit_packed: A uint8 numpy array with shape (num_shots, math.ceil(num_dets / 8)).
            Contains bit packed detection event data.
        decoder: The decoder to use for decoding, e.g. "pymatching".
        custom_decoders: Custom decoders that can be used if requested by name.
            If not specified, only decoders built into sinter, such as
            'pymatching' and 'fusion_blossom', can be used.

    Returns:
        A numpy uint8 array with shape (num_shots, math.ceil(num_obs / 8)).
        Contains bit packed observable prediction data.
    """

    decode_obj: Optional[Decoder] = None
    if custom_decoders is not None:
        decode_obj = custom_decoders.get(decoder)
    if decode_obj is None:
        decode_obj = BUILT_IN_DECODERS.get(decoder)
    if decode_obj is None:
        raise NotImplementedError(f"Unrecognized decoder: {decoder!r}")

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        dets_b8_path = tmp_dir / 'sinter_dets.b8'
        pred_b8_path = tmp_dir / 'sinter_predictions.b8'
        dem_path = tmp_dir / 'dem.dem'
        dem.to_file(dem_path)
        num_dets = dem.num_detectors
        num_obs = dem.num_observables

        stim.write_shot_data_file(
            data=dets_bit_packed,
            path=str(dets_b8_path),
            format='b8',
            num_detectors=num_dets,
            num_observables=0,
        )

        decode_obj.decode_via_files(
            num_shots=dets_bit_packed.shape[0],
            num_dets=num_dets,
            num_obs=num_obs,
            dem_path=dem_path,
            dets_b8_in_path=dets_b8_path,
            obs_predictions_b8_out_path=pred_b8_path,
            tmp_dir=tmp_dir,
        )

        return stim.read_shot_data_file(
            path=str(pred_b8_path),
            format='b8',
            bit_pack=True,
            num_detectors=0,
            num_observables=num_obs,
        )



import threading
from typing import List, Any

import sys
import time


class ThrottledProgressPrinter:
    """Handles printing progress updates interspersed amongst output.

    Throttles the progress updates to not flood the screen when 100 show up
    at the same time, and instead only show the latest one.
    """
    def __init__(self, *, outs: List[Any], print_progress: bool, min_progress_delay: float):
        self.outs = outs
        self.print_progress = print_progress
        self.next_can_print_time = time.monotonic()
        self.latest_msg = ''
        self.latest_printed_msg = ''
        self.min_progress_delay = min_progress_delay
        self.is_worker_running = False
        self.lock = threading.Lock()

    def print_out(self, msg: str) -> None:
        with self.lock:
            for out in self.outs:
                print(msg, file=out, flush=True)

    def show_latest_progress(self, msg: str) -> None:
        if not self.print_progress:
            return
        with self.lock:
            if msg == self.latest_msg:
                return
            self.latest_msg = msg
            if not self.is_worker_running:
                dt = self._try_print_else_delay()
                if dt > 0:
                    self.is_worker_running = True
                    threading.Thread(target=self._print_worker).start()

    def flush(self):
        with self.lock:
            if self.latest_msg != "" and self.latest_printed_msg != self.latest_msg:
                print('\033[31m' + self.latest_msg + '\033[0m', file=sys.stderr, flush=True)
                self.latest_printed_msg = self.latest_msg

    def _try_print_else_delay(self) -> float:
        t = time.monotonic()
        dt = self.next_can_print_time - t
        if dt <= 0:
            self.next_can_print_time = t + self.min_progress_delay
            self.is_worker_running = False
            if self.latest_msg != "" and self.latest_msg != self.latest_printed_msg:
                print('\033[31m' + self.latest_msg + '\033[0m', file=sys.stderr, flush=True)
                self.latest_printed_msg = self.latest_msg
        return max(dt, 0)

    def _print_worker(self):
        while True:
            with self.lock:
                dt = self._try_print_else_delay()
            if dt == 0:
                break
            time.sleep(dt)



from sinter._anon_task_stats import (
    AnonTaskStats,
)
from sinter._collection import (
    collect,
    iter_collect,
    post_selection_mask_from_4th_coord,
    Progress,
)
from sinter._collection_options import (
    CollectionOptions,
)
from sinter._csv_out import (
    CSV_HEADER,
)
from sinter._existing_data import (
    stats_from_csv_files,
)
from sinter._probability_util import (
    Fit,
    fit_binomial,
    fit_line_slope,
    fit_line_y_at_x,
    shot_error_rate_to_piece_error_rate,
    comma_separated_key_values,
)
from sinter._plotting import (
    better_sorted_str_terms,
    plot_discard_rate,
    plot_error_rate,
    group_by,
)
from sinter._task import (
    Task,
)
from sinter._task_stats import (
    TaskStats,
)
from sinter._predict import (
    predict_discards_bit_packed,
    predict_observables_bit_packed,
    predict_on_disk,
    predict_observables,
)
from sinter._decoding_decoder_class import (
    Decoder,
)



import math
from typing import Union

import numpy as np
import pytest

from sinter._probability_util import (
    binary_search, log_binomial, log_factorial, fit_line_y_at_x, fit_line_slope,
    binary_intercept, least_squares_through_point, fit_binomial, shot_error_rate_to_piece_error_rate,
)
from sinter._probability_util import comma_separated_key_values


@pytest.mark.parametrize(
    "arg,result",
    {
        0: 0,
        1: 0,
        2: math.log(2),
        3: math.log(2) + math.log(3),
        # These values were taken from wolfram alpha:
        10: 15.1044125730755152952257093292510,
        100: 363.73937555556349014407999336965,
        1000: 5912.128178488163348878130886725,
        10000: 82108.9278368143534553850300635,
        100000: 1051299.2218991218651292781082,
    }.items(),
)
def test_log_factorial(arg, result):
    np.testing.assert_allclose(log_factorial(arg), result, rtol=1e-5)


@pytest.mark.parametrize(
    "n,p,hits,result",
    [
        (1, 0.5, 0, np.log(0.5)),
        (1, 0.5, 1, np.log(0.5)),
        (1, 0.1, 0, np.log(0.9)),
        (1, 0.1, 1, np.log(0.1)),
        (2, [0, 1, 0.1, 0.5], 0, [0, -np.inf, np.log(0.9 ** 2), np.log(0.25)]),
        (2, [0, 1, 0.1, 0.5], 1, [-np.inf, -np.inf, np.log(0.1 * 0.9 * 2), np.log(0.5)]),
        (2, [0, 1, 0.1, 0.5], 2, [-np.inf, 0, np.log(0.1 ** 2), np.log(0.25)]),
        # Magic number comes from PDF[BinomialDistribution[10^10, 10^-6], 10000] on wolfram alpha.
        (10 ** 10, 10 ** -6, 10 ** 4, np.log(0.0039893915536591)),
        # Corner cases.
        (1, 0.0, 0, 0),
        (1, 0.0, 1, -np.inf),
        (1, 1.0, 0, -np.inf),
        (1, 1.0, 1, 0),
        # Array broadcast.
        (2, np.array([0.0, 0.5, 1.0]), 0, np.array([0.0, np.log(0.25), -np.inf])),
        (2, np.array([0.0, 0.5, 1.0]), 1, np.array([-np.inf, np.log(0.5), -np.inf])),
        (2, np.array([0.0, 0.5, 1.0]), 2, np.array([-np.inf, np.log(0.25), 0.0])),
    ],
)
def test_log_binomial(
    n: int, p: Union[float, np.ndarray], hits: int, result: Union[float, np.ndarray]
) -> None:
    np.testing.assert_allclose(log_binomial(n=n, p=p, hits=hits), result, rtol=1e-2)


def test_binary_search():
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=100.1) == 10
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=100) == 10
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=99.9) == 10
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=90) == 9
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=92) == 10
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=-100) == 0
    assert binary_search(func=lambda x: x**2, min_x=0, max_x=10**100, target=10**300) == 10**100


def test_least_squares_through_point():
    fit = least_squares_through_point(
        xs=np.array([1, 2, 3]),
        ys=np.array([2, 3, 4]),
        required_x=1,
        required_y=2)
    np.testing.assert_allclose(fit.slope, 1)
    np.testing.assert_allclose(fit.intercept, 1)

    fit = least_squares_through_point(
        xs=np.array([1, 2, 3]),
        ys=np.array([2, 3, 4]),
        required_x=1,
        required_y=1)
    np.testing.assert_allclose(fit.slope, 1.6, rtol=1e-5)
    np.testing.assert_allclose(fit.intercept, -0.6, atol=1e-5)


def test_binary_intercept():
    t = binary_intercept(func=lambda x: x**2, start_x=5, step=1, target_y=82.3, atol=0.01)
    assert t > 0 and abs(t**2 - 82.3) <= 0.01
    t = binary_intercept(func=lambda x: -x**2, start_x=5, step=1, target_y=-82.3, atol=0.01)
    assert t > 0 and abs(t**2 - 82.3) <= 0.01
    t = binary_intercept(func=lambda x: x**2, start_x=0, step=-1, target_y=82.3, atol=0.01)
    assert t < 0 and abs(t**2 - 82.3) <= 0.01
    t = binary_intercept(func=lambda x: -x**2, start_x=0, step=-1, target_y=-82.3, atol=0.2)
    assert t < 0 and abs(t**2 - 82.3) <= 0.2


def test_fit_y_at_x():
    fit = fit_line_y_at_x(
        xs=[1, 2, 3],
        ys=[1, 5, 9],
        target_x=100,
        max_extra_squared_error=1,
    )
    assert 300 < fit.low < 390 < fit.best < 410 < fit.high < 500


def test_fit_slope():
    fit = fit_line_slope(
        xs=[1, 2, 3],
        ys=[1, 5, 9],
        max_extra_squared_error=1,
    )
    np.testing.assert_allclose(fit.best, 4)
    assert 3 < fit.low < 3.5 < fit.best < 4.5 < fit.high < 5


def test_fit_binomial_shrink_towards_half():
    with pytest.raises(ValueError, match='max_likelihood_factor'):
        fit_binomial(num_shots=10 ** 5, num_hits=10 ** 5 / 2, max_likelihood_factor=0.1)

    fit = fit_binomial(num_shots=10 ** 5, num_hits=10 ** 5 / 2, max_likelihood_factor=1e3)
    np.testing.assert_allclose(
        (fit.low, fit.best, fit.high),
        (0.494122, 0.5, 0.505878),
        rtol=1e-4,
    )
    fit = fit_binomial(num_shots=10 ** 4, num_hits=10 ** 4 / 2, max_likelihood_factor=1e3)
    np.testing.assert_allclose(
        (fit.low, fit.best, fit.high),
        (0.481422, 0.5, 0.518578),
        rtol=1e-4,
    )
    fit = fit_binomial(num_shots=10 ** 4, num_hits=10 ** 4 / 2, max_likelihood_factor=1e2)
    np.testing.assert_allclose(
        (fit.low, fit.best, fit.high),
        (0.48483, 0.5, 0.51517),
        rtol=1e-4,
    )
    fit = fit_binomial(num_shots=1000, num_hits=500, max_likelihood_factor=1e3)
    np.testing.assert_allclose(
        (fit.low, fit.best, fit.high),
        (0.44143, 0.5, 0.55857),
        rtol=1e-4,
    )
    fit = fit_binomial(num_shots=100, num_hits=50, max_likelihood_factor=1e3)
    np.testing.assert_allclose(
        (fit.low, fit.best, fit.high),
        (0.3204, 0.5, 0.6796),
        rtol=1e-4,
    )


@pytest.mark.parametrize("n,c,factor", [
    (100, 50, 1e1),
    (100, 50, 1e2),
    (100, 50, 1e3),
    (1000, 500, 1e3),
    (10**6, 100, 1e3),
    (10**6, 100, 1e2),
])
def test_fit_binomial_vs_log_binomial(n: int, c: int, factor: float):
    fit = fit_binomial(num_shots=n, num_hits=n - c, max_likelihood_factor=factor)
    a = fit.low
    b = fit.high

    raw = log_binomial(p=(n - c) / n, n=n, hits=n - c)
    low = log_binomial(p=a, n=n, hits=n - c)
    high = log_binomial(p=b, n=n, hits=n - c)

    np.testing.assert_allclose(
        fit.best,
        (n - c) / n,
        rtol=1e-4,
    )

    np.testing.assert_allclose(
        np.exp(raw - low),
        factor,
        rtol=1e-2,
    )
    np.testing.assert_allclose(
        np.exp(raw - high),
        factor,
        rtol=1e-2,
    )


def test_comma_separated_key_values():
    d = comma_separated_key_values("folder/a=2,b=3.0,c=test.stim")
    assert d == {
        'a': 2,
        'b': 3.0,
        'c': 'test',
    }
    assert type(d['a']) == int
    assert type(d['b']) == float
    with pytest.raises(ValueError, match='separated'):
        comma_separated_key_values("folder/a,b=3.0,c=test.stim")


def test_shot_error_rate_to_piece_error_rate():
    np.testing.assert_allclose(
        shot_error_rate_to_piece_error_rate(
            shot_error_rate=0.2 * (1 - 0.2) * 2,
            pieces=2,
        ),
        0.2,
        rtol=1e-5)

    np.testing.assert_allclose(
        shot_error_rate_to_piece_error_rate(
            shot_error_rate=0.001 * (1 - 0.001) * 2,
            pieces=2,
        ),
        0.001,
        rtol=1e-5)

    np.testing.assert_allclose(
        shot_error_rate_to_piece_error_rate(
            shot_error_rate=0.001 * (1 - 0.001)**2 * 3 + 0.001**3,
            pieces=3,
        ),
        0.001,
        rtol=1e-5)

    # Extremely low error rates.
    np.testing.assert_allclose(
        shot_error_rate_to_piece_error_rate(
            shot_error_rate=1e-100,
            pieces=100,
        ),
        1e-102,
        rtol=1e-5)



import pathlib


def decode_using_pymatching_v2(*,
                               num_shots: int,
                               num_dets: int,
                               num_obs: int,
                               dem_path: pathlib.Path,
                               dets_b8_in_path: pathlib.Path,
                               obs_predictions_b8_out_path: pathlib.Path,
                               tmp_dir: pathlib.Path,
                               ) -> None:
    """Use pymatching to predict observables from detection events."""

    import pymatching
    if num_dets == 0:
        with open(obs_predictions_b8_out_path, 'wb') as f:
            f.write(b'\0' * (num_obs * num_shots))
        return

    result = pymatching._cpp_pymatching.main(command_line_args=[
        "predict",
        "--dem", str(dem_path),
        "--in", str(dets_b8_in_path),
        "--in_format", "b8",
        "--out", str(obs_predictions_b8_out_path),
        "--out_format", "b8",
    ])
    if result:
        raise ValueError("pymatching.main returned a non-zero exit code")



import dataclasses
import math
import pathlib
import stim
from typing import Any
from typing import Dict
from typing import Union, Callable, Sequence, Tuple, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import sinter
    from scipy.stats._stats_mstats_common import LinregressResult



def log_binomial(*, p: Union[float, np.ndarray], n: int, hits: int) -> Union[float, np.ndarray]:
    r"""Approximates $\ln(P(hits = B(n, p)))$; the natural logarithm of a binomial distribution.

    All computations are done in log space to ensure intermediate values can be represented as
    floating point numbers without underflowing to 0 or overflowing to infinity. This is necessary
    when computing likelihoods over many samples. For example, if 80% of a million samples are hits,
    the maximum likelihood estimate is p=0.8. But even this optimal estimate assigns a prior
    probability of roughly 10^-217322 for seeing *exactly* 80% hits out of a million (whereas the
    smallest representable double is roughly 10^-324).

    This method can be broadcast over multiple hypothesis probabilities.

    Args:
        p: The independent probability of a hit occurring for each sample. This can also be an array
            of probabilities, in which case the function is broadcast over the array.
        n: The number of samples that were taken.
        hits: The number of hits that were observed amongst the samples that were taken.

    Returns:
        $\ln(P(hits = B(n, p)))$
    """
    # Clamp probabilities into the valid [0, 1] range (in case float error put them outside it).
    p_clipped = np.clip(p, 0, 1)

    result = np.zeros(shape=p_clipped.shape, dtype=np.float32)
    misses = n - hits

    # Handle p=0 and p=1 cases separately, to avoid arithmetic warnings.
    if hits:
        result[p_clipped == 0] = -np.inf
    if misses:
        result[p_clipped == 1] = -np.inf

    # Multiply p**hits and (1-p)**misses onto the total, in log space.
    result[p_clipped != 0] += np.log(p_clipped[p_clipped != 0]) * hits
    result[p_clipped != 1] += np.log1p(-p_clipped[p_clipped != 1]) * misses

    # Multiply (n choose hits) onto the total, in log space.
    log_n_choose_hits = log_factorial(n) - log_factorial(misses) - log_factorial(hits)
    result += log_n_choose_hits

    return result


def log_factorial(n: int) -> float:
    r"""Approximates $\ln(n!)$; the natural logarithm of a factorial.

    Uses Stirling's approximation for large n.
    """
    if n < 20:
        return sum(math.log(k) for k in range(1, n + 1))
    return (n + 0.5) * math.log(n) - n + math.log(2 * np.pi) / 2


def binary_search(*, func: Callable[[int], float], min_x: int, max_x: int, target: float) -> int:
    """Performs an approximate granular binary search over a monotonically ascending function."""
    while max_x > min_x + 1:
        med_x = (min_x + max_x) // 2
        out = func(med_x)
        if out < target:
            min_x = med_x
        elif out > target:
            max_x = med_x
        else:
            return med_x
    fmax = func(max_x)
    fmin = func(min_x)
    dmax = 0 if fmax == target else fmax - target
    dmin = 0 if fmin == target else fmin - target
    return max_x if abs(dmax) < abs(dmin) else min_x


def binary_intercept(*, func: Callable[[float], float], start_x: float, step: float, target_y: float, atol: float) -> float:
    """Performs an approximate granular binary search over a monotonically ascending function."""
    start_y = func(start_x)
    if abs(start_y - target_y) <= atol:
        return start_x
    while (func(start_x + step) >= target_y) == (start_y >= target_y):
        step *= 2
        if np.isinf(step) or step == 0:
            raise ValueError("Failed.")
    xs = [start_x, start_x + step]
    min_x = min(xs)
    max_x = max(xs)
    increasing = func(min_x) < func(max_x)

    while True:
        med_x = (min_x + max_x) / 2
        med_y = func(med_x)
        if abs(med_y - target_y) <= atol:
            return med_x
        assert med_x not in [min_x, max_x]
        if (med_y < target_y) == increasing:
            min_x = med_x
        else:
            max_x = med_x


def least_squares_cost(*, xs: np.ndarray, ys: np.ndarray, intercept: float, slope: float) -> float:
    assert len(xs.shape) == 1
    assert xs.shape == ys.shape
    return np.sum((intercept + slope*xs - ys)**2)


def least_squares_through_point(*, xs: np.ndarray, ys: np.ndarray, required_x: float, required_y: float) -> 'LinregressResult':
    # Local import to reduce initial cost of importing sinter.
    from scipy.optimize import leastsq
    from scipy.stats._stats_mstats_common import LinregressResult

    xs2 = xs - required_x
    ys2 = ys - required_y

    def err(slope: float) -> float:
        return least_squares_cost(xs=xs2, ys=ys2, intercept=0, slope=slope)

    (best_slope,), _ = leastsq(func=err, x0=0.0)
    intercept = required_y - required_x * best_slope
    return LinregressResult(best_slope, intercept, None, None, None, intercept_stderr=False)


def least_squares_with_slope(*, xs: np.ndarray, ys: np.ndarray, required_slope: float) -> 'LinregressResult':
    def err(intercept: float) -> float:
        return least_squares_cost(xs=xs, ys=ys, intercept=intercept, slope=required_slope)

    # Local import to reduce initial cost of importing sinter.
    from scipy.optimize import leastsq
    from scipy.stats._stats_mstats_common import LinregressResult
    (best_intercept,), _ = leastsq(func=err, x0=0.0)
    return LinregressResult(required_slope, best_intercept, None, None, None, intercept_stderr=False)


@dataclasses.dataclass(frozen=True)
class Fit:
    """The result of a fitting process.

    Attributes:
        best: The max likelihood hypothesis. The hypothesis that had the lowest squared error,
            or the best fitting score.
        low: The hypothesis with the smallest parameter whose cost or score was still "close to"
            the cost of the best hypothesis. For example, this could be a hypothesis whose
            squared error was within some tolerance of the best fit's square error, or whose
            likelihood was within some maximum Bayes factor of the max likelihood hypothesis.
        high: The hypothesis with the larger parameter whose cost or score was still "close to"
            the cost of the best hypothesis. For example, this could be a hypothesis whose
            squared error was within some tolerance of the best fit's square error, or whose
            likelihood was within some maximum Bayes factor of the max likelihood hypothesis.
    """
    low: float
    best: float
    high: float


def fit_line_y_at_x(*,
                    xs: Sequence[float],
                    ys: Sequence[float],
                    target_x: float,
                    max_extra_squared_error: float) -> 'sinter.Fit':
    """Performs a line fit of the given points, focusing on the line's value for y at a given x coordinate.

    Finds the y value at the given x of the best fit, but also the minimum and maximum
    values for y at the given x amongst all possible line fits whose squared error cost
    is within the given `max_extra_squared_error` cost of the best fit.

    Args:
        xs: The x coordinates of points to fit.
        ys: The y coordinates of points to fit.
        target_x: The reported fit values are the value of y at this x coordinate.
        max_extra_squared_error: When computing the low and high fits, this is
            the maximum additional squared error that can be introduced by
            varying the slope away from the best fit.

    Returns:
        A sinter.Fit containing the best fit for y at the given x, as well as low
        and high fits that are as far as possible from the best fit while respective
        the given max_extra_squared_error.
    """

    # Local import to reduce initial cost of importing sinter.
    from scipy.stats import linregress

    xs = np.array(xs, dtype=np.float64)
    ys = np.array(ys, dtype=np.float64)
    fit = linregress(xs, ys)
    base_cost = least_squares_cost(xs=xs, ys=ys, intercept=fit.intercept, slope=fit.slope)
    base_y = float(fit.intercept + target_x * fit.slope)

    def cost_for_y(y2: float) -> float:
        fit2 = least_squares_through_point(xs=xs, ys=ys, required_x=target_x, required_y=y2)
        return least_squares_cost(xs=xs, ys=ys, intercept=fit2.intercept, slope=fit2.slope)

    low_y = binary_intercept(start_x=base_y, step=-1, target_y=base_cost + max_extra_squared_error, func=cost_for_y, atol=1e-5)
    high_y = binary_intercept(start_x=base_y, step=1, target_y=base_cost + max_extra_squared_error, func=cost_for_y, atol=1e-5)
    return Fit(low=low_y, best=base_y, high=high_y)


def fit_line_slope(*,
              xs: Sequence[float],
              ys: Sequence[float],
              max_extra_squared_error: float) -> 'sinter.Fit':
    """Performs a line fit of the given points, focusing on the line's slope.

    Finds the slope of the best fit, but also the minimum and maximum slopes
    for line fits whose squared error cost is within the given `max_extra_squared_error`
    cost of the best fit.

    Args:
        xs: The x coordinates of points to fit.
        ys: The y coordinates of points to fit.
        max_extra_squared_error: When computing the low and high fits, this is
            the maximum additional squared error that can be introduced by
            varying the slope away from the best fit.

    Returns:
        A sinter.Fit containing the best fit, as well as low and high fits that
        are as far as possible from the best fit while respective the given
        max_extra_squared_error.
    """
    # Local import to reduce initial cost of importing sinter.
    from scipy.stats import linregress

    xs = np.array(xs, dtype=np.float64)
    ys = np.array(ys, dtype=np.float64)
    fit = linregress(xs, ys)
    base_cost = least_squares_cost(xs=xs, ys=ys, intercept=fit.intercept, slope=fit.slope)

    def cost_for_slope(slope: float) -> float:
        fit2 = least_squares_with_slope(xs=xs, ys=ys, required_slope=slope)
        return least_squares_cost(xs=xs, ys=ys, intercept=fit2.intercept, slope=fit2.slope)

    low_slope = binary_intercept(start_x=fit.slope, step=-1, target_y=base_cost + max_extra_squared_error, func=cost_for_slope, atol=1e-5)
    high_slope = binary_intercept(start_x=fit.slope, step=1, target_y=base_cost + max_extra_squared_error, func=cost_for_slope, atol=1e-5)
    return Fit(low=low_slope, best=fit.slope, high=high_slope)


def fit_binomial(
        *,
        num_shots: int,
        num_hits: int,
        max_likelihood_factor: float) -> 'sinter.Fit':
    """Compute the (min, max) probabilities within the given ratio of the max likelihood hypothesis.

    Args:
        num_shots: The number of samples that were taken.
        num_hits: The number of hits that were seen in the samples.
        max_likelihood_factor: The maximum Bayes factor between the low,high hypotheses and
            the best hypothesis (the max likelihood hypothesis). This value should be
            larger than 1 (as opposed to between 0 and 1).

    Returns:
        A `sinter.Fit` containing the max likelihood h
        A (min_hypothesis_probability, max_hypothesis_probability) tuple.
    """
    if max_likelihood_factor < 1:
        raise ValueError(f'max_likelihood_factor={max_likelihood_factor} < 1')
    if num_shots == 0:
        return Fit(low=0, high=1, best=0.5)
    log_max_likelihood = log_binomial(p=num_hits / num_shots, n=num_shots, hits=num_hits)
    target_log_likelihood = log_max_likelihood - math.log(max_likelihood_factor)
    acc = 100
    low = binary_search(
        func=lambda exp_err: log_binomial(p=exp_err / (acc * num_shots), n=num_shots, hits=num_hits),
        target=target_log_likelihood,
        min_x=0,
        max_x=num_hits * acc) / acc
    high = binary_search(
        func=lambda exp_err: -log_binomial(p=exp_err / (acc * num_shots), n=num_shots, hits=num_hits),
        target=-target_log_likelihood,
        min_x=num_hits * acc,
        max_x=num_shots * acc) / acc
    return Fit(best=num_hits / num_shots, low=low / num_shots, high=high / num_shots)


def shot_error_rate_to_piece_error_rate(shot_error_rate: float, *, pieces: float) -> float:
    """Convert from total error rate to per-piece error rate.

    Works by assuming pieces fail independently and a shot fails if an any single
    piece fails.
    """
    if not (0 <= shot_error_rate <= 1):
        raise ValueError(f'need (0 <= shot_error_rate={shot_error_rate} <= 1)')
    if pieces <= 0:
        raise ValueError('need pieces > 0')
    if not isinstance(pieces, (int, float)):
        raise ValueError('need isinstance(pieces, (int, float)')
    if pieces == 1:
        return shot_error_rate
    if shot_error_rate > 0.5:
        return 1 - shot_error_rate_to_piece_error_rate(1 - shot_error_rate, pieces=pieces)
    assert 0 <= shot_error_rate <= 0.5
    randomize_rate = 2*shot_error_rate
    round_randomize_rate = 1 - (1 - randomize_rate)**(1 / pieces)
    round_error_rate = round_randomize_rate / 2

    if round_error_rate == 0:
        # The intermediate numbers got too small. Fallback to division approximation.
        return shot_error_rate / pieces

    return round_error_rate


def comma_separated_key_values(path: str) -> Dict[str, Any]:
    name = pathlib.Path(path).name
    if '.' in name:
        name = name[:name.rindex('.')]
    result = {}
    for term in name.split(','):
        parts = term.split('=')
        if len(parts) != 2:
            raise ValueError(f"Expected a path with a filename containing comma-separated key=value terms like 'a=2,b=3.stim', but got {path!r}.")
        k, v = parts
        try:
            v = int(v)
        except ValueError:
            try:
                v = float(v)
            except ValueError:
                pass
        result[k] = v
    return result



import dataclasses
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    import sinter


@dataclasses.dataclass(frozen=True)
class CollectionOptions:
    """Configures how data is collected for a decoding problem.

    Attributes:
        max_shots: Defaults to None (unused). Stops the sampling process
            after this many samples have been taken from the circuit.
        max_errors: Defaults to None (unused). Stops the sampling process
            after this many errors have been seen in samples taken from the
            circuit. The actual number sampled errors may be larger due to
            batching.
        start_batch_size: Defaults to None (collector's choice). The very
            first shots taken from the circuit will use a batch of this
            size, and no other batches will be taken in parallel. Once this
            initial fact finding batch is done, batches can be taken in
            parallel and the normal batch size limiting processes take over.
        max_batch_size: Defaults to None (unused). Limits batches from
            taking more than this many shots at once. For example, this can
            be used to ensure memory usage stays below some limit.
        max_batch_seconds: Defaults to None (unused). When set, the recorded
            data from previous shots is used to estimate how much time is
            taken per shot. This information is then used to predict the
            biggest batch size that can finish in under the given number of
            seconds. Limits each batch to be no larger than that.
    """

    max_shots: Optional[int] = None
    max_errors: Optional[int] = None
    start_batch_size: Optional[int] = None
    max_batch_size: Optional[int] = None
    max_batch_seconds: Optional[float] = None

    def __post_init__(self):
        if self.max_shots is not None and self.max_shots < 0:
            raise ValueError(f'max_shots is not None and max_shots={self.max_shots} < 0')
        if self.max_errors is not None and self.max_errors < 0:
            raise ValueError(f'max_errors is not None and max_errors={self.max_errors} < 0')
        if self.start_batch_size is not None and self.start_batch_size <= 0:
            raise ValueError(f'start_batch_size is not None and start_batch_size={self.start_batch_size} <= 0')
        if self.max_batch_size is not None and self.max_batch_size <= 0:
            raise ValueError(
                f'max_batch_size={self.max_batch_size} is not None and max_batch_size <= 0')
        if self.max_batch_seconds is not None and self.max_batch_seconds <= 0:
            raise ValueError(
                f'max_batch_seconds={self.max_batch_seconds} is not None and max_batch_seconds <= 0')

    def combine(self, other: 'sinter.CollectionOptions') -> 'sinter.CollectionOptions':
        return CollectionOptions(
            max_shots=nullable_min(self.max_shots, other.max_shots),
            max_errors=nullable_min(self.max_errors, other.max_errors),
            start_batch_size=nullable_min(self.start_batch_size, other.start_batch_size),
            max_batch_size=nullable_min(self.max_batch_size, other.max_batch_size),
            max_batch_seconds=nullable_min(self.max_batch_seconds, other.max_batch_seconds))


def nullable_min(a: Optional[int], b: Optional[int]) -> Optional[int]:
    if a is None:
        return b
    if b is None:
        return a
    return min(a, b)



import pathlib

from sinter._decoding_decoder_class import Decoder


class InternalDecoder(Decoder):
    """Use internal decoders to predict observables from detection events."""

    def __init__(self, name: str):
        self.name = name

    def decode_via_files(self,
                         *,
                         num_shots: int,
                         num_dets: int,
                         num_obs: int,
                         dem_path: pathlib.Path,
                         dets_b8_in_path: pathlib.Path,
                         obs_predictions_b8_out_path: pathlib.Path,
                         tmp_dir: pathlib.Path,
                       ) -> None:

        if num_dets == 0:
            with open(obs_predictions_b8_out_path, 'wb') as f:
                f.write(b'\0' * (num_obs * num_shots))
            return

        try:
            import gqec  # Internal python wheel.
        except ImportError as ex:
            raise ImportError(
                "The decoder 'internal*' isn't installed.\n"
                "These decoders aren't publicly available.\n"
            ) from ex

        gqec.run_for_sinter(str(dem_path), str(dets_b8_in_path), "b8", str(obs_predictions_b8_out_path), self.name)



import pathlib

from sinter._decoding_decoder_class import Decoder
from sinter._decoding_pymatching_v1 import decode_using_pymatching_v1
from sinter._decoding_pymatching_v2 import decode_using_pymatching_v2


class PyMatchingDecoder(Decoder):
    """Use pymatching to predict observables from detection events."""

    def __init__(self):
        self._method_to_use = None

    def decode_via_files(self,
                         *,
                         num_shots: int,
                         num_dets: int,
                         num_obs: int,
                         dem_path: pathlib.Path,
                         dets_b8_in_path: pathlib.Path,
                         obs_predictions_b8_out_path: pathlib.Path,
                         tmp_dir: pathlib.Path,
                       ) -> None:
        try:
            import pymatching
        except ImportError as ex:
            raise ImportError(
                "The decoder 'pymatching' isn't installed\n"
                "To fix this, install the python package 'pymatching' into your environment.\n"
                "For example, if you are using pip, run `pip install pymatching`.\n"
            ) from ex

        if self._method_to_use is None:
            if getattr(getattr(pymatching, '_cpp_pymatching', None), 'main', None) is not None:
                self._method_to_use = decode_using_pymatching_v2
            else:
                self._method_to_use = decode_using_pymatching_v1

        self._method_to_use(
            num_shots=num_shots,
            num_dets=num_dets,
            num_obs=num_obs,
            dem_path=dem_path,
            dets_b8_in_path=dets_b8_in_path,
            obs_predictions_b8_out_path=obs_predictions_b8_out_path,
            tmp_dir=tmp_dir,
        )



import csv
import io
import json
from typing import Any, Optional

from sinter._json_type import JSON_TYPE


def escape_csv(text: Any, width: Optional[int]) -> str:
    output = io.StringIO()
    csv.writer(output).writerow([text])
    text = output.getvalue().strip()
    if width is not None:
        text = text.rjust(width)
    return text


def csv_line(*,
             shots: Any,
             errors: Any,
             discards: Any,
             seconds: Any,
             decoder: Any,
             strong_id: Any,
             json_metadata: JSON_TYPE,
             is_header: bool = False) -> str:
    if isinstance(seconds, float):
        if seconds < 1:
            seconds = f'{seconds:0.3f}'
        elif seconds < 10:
            seconds = f'{seconds:0.2f}'
        else:
            seconds = f'{seconds:0.1f}'
    if not is_header:
        json_metadata = json.dumps(json_metadata,
                                   separators=(',', ':'),
                                   sort_keys=True)

    shots = escape_csv(shots, 10)
    errors = escape_csv(errors, 10)
    discards = escape_csv(discards, 10)
    seconds = escape_csv(seconds, 8)
    decoder = escape_csv(decoder, None)
    strong_id = escape_csv(strong_id, None)
    json_metadata = escape_csv(json_metadata, None)
    return (f'{shots},'
            f'{errors},'
            f'{discards},'
            f'{seconds},'
            f'{decoder},'
            f'{strong_id},'
            f'{json_metadata}')


CSV_HEADER = csv_line(shots='shots',
                      errors='errors',
                      discards='discards',
                      seconds='seconds',
                      strong_id='strong_id',
                      decoder='decoder',
                      json_metadata='json_metadata',
                      is_header=True)



from typing import Dict

from sinter._decoding_decoder_class import Decoder
from sinter._decoding_fusion_blossom import FusionBlossomDecoder
from sinter._decoding_internal import InternalDecoder
from sinter._decoding_pymatching import PyMatchingDecoder

BUILT_IN_DECODERS: Dict[str, Decoder] = {
    'pymatching': PyMatchingDecoder(),
    'fusion_blossom': FusionBlossomDecoder(),
    'internal': InternalDecoder('internal'),
    'internal_correlated': InternalDecoder('internal_correlated'),
    'internal_2': InternalDecoder('internal_2'),
    'internal_correlated_2': InternalDecoder('internal_correlated_2'),
}



from typing import Any, Dict, Optional, TYPE_CHECKING, Iterable

import hashlib
import json
import math
import numpy as np

from sinter._collection_options import CollectionOptions
from sinter._json_type import JSON_TYPE

if TYPE_CHECKING:
    import sinter
    import stim


class Task:
    """A decoding problem that can be sampled from."""

    def __init__(
        self,
        *,
        circuit: 'stim.Circuit',
        decoder: Optional[str] = None,
        detector_error_model: Optional['stim.DetectorErrorModel'] = None,
        postselection_mask: Optional[np.ndarray] = None,
        postselected_observables_mask: Optional[np.ndarray] = None,
        json_metadata: JSON_TYPE = None,
        collection_options: 'sinter.CollectionOptions' = CollectionOptions(),
        skip_validation: bool = False,
        _unvalidated_strong_id: Optional[str] = None,
    ) -> None:
        """
        Args:
            circuit: The annotated noisy circuit to sample detection event data
                and logical observable data form.
            decoder: The decoder to use to predict the logical observable data
                from the detection event data. This can be set to None if it
                will be specified later (e.g. by the call to `collect`).
            detector_error_model: Specifies the error model to give to the decoder.
                Defaults to None, indicating that it should be automatically derived
                using `stim.Circuit.detector_error_model`.
            postselection_mask: Defaults to None (unused). A bit packed bitmask
                identifying detectors that must not fire. Shots where the
                indicated detectors fire are discarded.
            postselected_observables_mask: Defaults to None (unused). A bit
                packed bitmask identifying observable indices to postselect on.
                Anytime the decoder's predicted flip for one of these
                observables doesn't agree with the actual measured flip value of
                the observable, the shot is discarded instead of counting as an
                error.
            json_metadata: Defaults to None. Custom additional data describing
                the problem. Must be JSON serializable. For example, this could
                be a dictionary with "physical_error_rate" and "code_distance"
                keys.
            collection_options: Specifies custom options for collecting this
                single task. These options are merged with the global options
                to determine what happens.

                For example, if a task has `collection_options` set to
                `sinter.CollectionOptions(max_shots=1000, max_errors=100)` and
                `sinter.collect` was called with `max_shots=500` and
                `max_errors=200`, then either 500 shots or 100 errors will be
                collected for the task (whichever comes first).
            skip_validation: Defaults to False. Normally the arguments given to
                this method are checked for consistency (e.g. the detector error
                model should have the same number of detectors as the circuit).
                Setting this argument to True will skip doing the consistency
                checks. Note that this can result in confusing errors later, if
                the arguments are not actually consistent.
            _unvalidated_strong_id: Must be set to None unless `skip_validation`
                is set to True. Otherwise, if this is specified then it should
                be equal to the value returned by self.strong_id().
        """
        if not skip_validation:
            if _unvalidated_strong_id is not None:
                raise ValueError("_unvalidated_strong_id is not None and not skip_validation")
            num_dets = circuit.num_detectors
            num_obs = circuit.num_observables
            dem = detector_error_model
            if dem is not None:
                if circuit.num_detectors != dem.num_detectors:
                    raise ValueError(f"circuit.num_detectors={num_dets!r} != detector_error_model.num_detectors={dem.num_detectors!r}")
                if num_obs != dem.num_observables:
                    raise ValueError(f"circuit.num_observables={num_obs!r} != detector_error_model.num_observables={dem.num_observables!r}")
            if postselection_mask is not None:
                if not isinstance(postselection_mask, np.ndarray):
                    raise ValueError(f"not isinstance(postselection_mask={postselection_mask!r}, np.ndarray)")
                if postselection_mask.dtype != np.uint8:
                    raise ValueError(f"postselection_mask.dtype={postselection_mask.dtype!r} != np.uint8")
                shape = (math.ceil(num_dets / 8),)
                if postselection_mask.shape != shape:
                    raise ValueError(f"postselection_mask.shape={postselection_mask.shape!r} != (math.ceil(circuit.num_detectors / 8),)={shape!r}")
            if postselected_observables_mask is not None:
                if not isinstance(postselected_observables_mask, np.ndarray):
                    raise ValueError(f"not isinstance(postselected_observables_mask={postselected_observables_mask!r}, np.ndarray)")
                if postselected_observables_mask.dtype != np.uint8:
                    raise ValueError(f"postselected_observables_mask.dtype={postselected_observables_mask.dtype!r} != np.uint8")
                shape = (math.ceil(num_obs / 8),)
                if postselected_observables_mask.shape != shape:
                    raise ValueError(f"postselected_observables_mask.shape={postselected_observables_mask.shape!r} != (math.ceil(circuit.num_observables / 8),)={shape!r}")
        self.circuit = circuit
        self.decoder = decoder
        self.detector_error_model = detector_error_model
        self.postselection_mask = postselection_mask
        self.postselected_observables_mask = postselected_observables_mask
        self.json_metadata = json_metadata
        self.collection_options = collection_options
        self._unvalidated_strong_id = _unvalidated_strong_id

    def strong_id_value(self) -> Dict[str, Any]:
        """Contains all raw values that affect the strong id.

        This value is converted into the actual strong id by:
            - Serializing it into text using JSON.
            - Serializing the JSON text into bytes using UTF8.
            - Hashing the UTF8 bytes using SHA256.
        """
        if self.decoder is None:
            raise ValueError("Can't compute strong_id until `decoder` is set.")
        if self.detector_error_model is None:
            raise ValueError("Can't compute strong_id until `detector_error_model` is set.")
        result = {
            'circuit': str(self.circuit),
            'decoder': self.decoder,
            'decoder_error_model': str(self.detector_error_model),
            'postselection_mask':
                None
                if self.postselection_mask is None
                else [int(e) for e in self.postselection_mask],
            'json_metadata': self.json_metadata,
        }
        if self.postselected_observables_mask is not None:
            result['postselected_observables_mask'] = [int(e) for e in self.postselected_observables_mask]
        return result

    def strong_id_text(self) -> str:
        """The text that is serialized and hashed to get the strong id.

        This value is converted into the actual strong id by:
            - Serializing into bytes using UTF8.
            - Hashing the UTF8 bytes using SHA256.
        """
        return json.dumps(self.strong_id_value())

    def strong_id_bytes(self) -> bytes:
        """The bytes that are hashed to get the strong id.

        This value is converted into the actual strong id by:
            - Hashing these bytes using SHA256.
        """
        return self.strong_id_text().encode('utf8')

    def _recomputed_strong_id(self) -> str:
        return hashlib.sha256(self.strong_id_bytes()).hexdigest()

    def strong_id(self) -> str:
        """A cryptographically unique identifier for this task.

        This value is affected by:
            - The exact circuit.
            - The exact detector error model.
            - The decoder.
            - The json metadata.
            - The postselection mask.
        """
        if self._unvalidated_strong_id is None:
            self._unvalidated_strong_id = self._recomputed_strong_id()
        return self._unvalidated_strong_id



from typing import Optional, Dict, Tuple, TYPE_CHECKING, Union

import contextlib
import pathlib
import tempfile
import math
import time

import numpy as np
import stim

from sinter._anon_task_stats import AnonTaskStats
from sinter._decoding_all_built_in_decoders import BUILT_IN_DECODERS
from sinter._decoding_decoder_class import Decoder

if TYPE_CHECKING:
    import sinter


def streaming_post_select(*,
                          num_dets: int,
                          num_obs: int,
                          dets_in_b8: pathlib.Path,
                          obs_in_b8: Optional[pathlib.Path],
                          dets_out_b8: pathlib.Path,
                          obs_out_b8: Optional[pathlib.Path],
                          discards_out_b8: Optional[pathlib.Path],
                          num_shots: int,
                          post_mask: np.ndarray) -> int:
    if post_mask.shape != ((num_dets + 7) // 8,):
        raise ValueError(f"post_mask.shape={post_mask.shape} != (math.ceil(num_detectors / 8),)")
    if post_mask.dtype != np.uint8:
        raise ValueError(f"post_mask.dtype={post_mask.dtype} != np.uint8")
    assert (obs_in_b8 is None) == (obs_out_b8 is None)

    num_det_bytes = math.ceil(num_dets / 8)
    num_obs_bytes = math.ceil(num_obs / 8)
    num_shots_left = num_shots
    num_discards = 0

    with contextlib.ExitStack() as ctx:
        dets_in_f = ctx.enter_context(open(dets_in_b8, 'rb'))
        dets_out_f = ctx.enter_context(open(dets_out_b8, 'wb'))
        if obs_in_b8 is not None and obs_out_b8 is not None:
            obs_in_f = ctx.enter_context(open(obs_in_b8, 'rb'))
            obs_out_f = ctx.enter_context(open(obs_out_b8, 'wb'))
        else:
            obs_in_f = None
            obs_out_f = None
        if discards_out_b8 is not None:
            discards_out_f = ctx.enter_context(open(discards_out_b8, 'wb'))
        else:
            discards_out_f = None

        while num_shots_left:
            batch_size = min(num_shots_left, math.ceil(10 ** 6 / max(1, num_dets)))

            det_batch = np.fromfile(dets_in_f, dtype=np.uint8, count=num_det_bytes * batch_size)
            det_batch.shape = (batch_size, num_det_bytes)
            discarded = np.any(det_batch & post_mask, axis=1)
            det_left = det_batch[~discarded, :]
            det_left.tofile(dets_out_f)

            if obs_in_f is not None and obs_out_f is not None:
                obs_batch = np.fromfile(obs_in_f, dtype=np.uint8, count=num_obs_bytes * batch_size)
                obs_batch.shape = (batch_size, num_obs_bytes)
                obs_left = obs_batch[~discarded, :]
                obs_left.tofile(obs_out_f)
            if discards_out_f is not None:
                discarded.tofile(discards_out_f)

            num_discards += np.count_nonzero(discarded)
            num_shots_left -= batch_size

    return num_discards


def _streaming_count_mistakes(
        *,
        num_shots: int,
        num_obs: int,
        postselected_observable_mask: Optional[np.ndarray] = None,
        obs_in: pathlib.Path,
        predictions_in: pathlib.Path) -> Tuple[int, int]:

    num_obs_bytes = math.ceil(num_obs / 8)
    num_shots_left = num_shots
    num_errors = 0
    num_discards = 0
    with open(obs_in, 'rb') as obs_in_f:
        with open(predictions_in, 'rb') as predictions_in_f:
            while num_shots_left:
                batch_size = min(num_shots_left, math.ceil(10**6 / max(num_obs, 1)))

                obs_batch = np.fromfile(obs_in_f, dtype=np.uint8, count=num_obs_bytes * batch_size)
                pred_batch = np.fromfile(predictions_in_f, dtype=np.uint8, count=num_obs_bytes * batch_size)
                obs_batch.shape = (batch_size, num_obs_bytes)
                pred_batch.shape = (batch_size, num_obs_bytes)

                cmp_table = pred_batch ^ obs_batch
                if postselected_observable_mask is None:
                    local_discards = 0
                else:
                    local_discards = np.count_nonzero(np.any(cmp_table & postselected_observable_mask, axis=1))
                local_errors_or_discards = np.count_nonzero(np.any(cmp_table, axis=1))
                num_discards += local_discards
                num_errors += local_errors_or_discards - local_discards
                num_shots_left -= batch_size
    return num_discards, num_errors


def sample_decode(*,
                  circuit_obj: Optional[stim.Circuit],
                  circuit_path: Union[None, str, pathlib.Path],
                  dem_obj: Optional[stim.DetectorErrorModel],
                  dem_path: Union[None, str, pathlib.Path],
                  post_mask: Optional[np.ndarray] = None,
                  postselected_observable_mask: Optional[np.ndarray] = None,
                  num_shots: int,
                  decoder: str,
                  tmp_dir: Union[str, pathlib.Path, None] = None,
                  custom_decoders: Optional[Dict[str, 'sinter.Decoder']] = None,
                  ) -> AnonTaskStats:
    """Samples how many times a decoder correctly predicts the logical frame.

    Args:
        circuit_obj: The noisy circuit to sample from and decode results for.
            Must specify circuit_obj XOR circuit_path.
        circuit_path: The file storing the circuit to sample from.
            Must specify circuit_obj XOR circuit_path.
        dem_obj: The error model to give to the decoder.
            Must specify dem_obj XOR dem_path.
        dem_path: The file storing the error model to give to the decoder.
            Must specify dem_obj XOR dem_path.
        post_mask: Postselection mask. Any samples that have a non-zero result
            at a location where the mask has a 1 bit are discarded. If set to
            None, no postselection is performed.
        postselected_observable_mask: Bit packed mask indicating which observables to
            postselect on. If the decoder incorrectly predicts any of these observables, the
            shot is discarded instead of counted as an error.
        num_shots: The number of sample shots to take from the circuit.
        decoder: The name of the decoder to use. Allowed values are:
            "pymatching":
                Use pymatching min-weight-perfect-match decoder.
            "internal":
                Use internal decoder with uncorrelated decoding.
            "internal_correlated":
                Use internal decoder with correlated decoding.
        tmp_dir: An existing directory that is currently empty where temporary
            files can be written as part of performing decoding. If set to
            None, one is created using the tempfile package.
        custom_decoders: Custom decoders that can be used if requested by name.
            If not specified, only decoders built into sinter, such as
            'pymatching' and 'fusion_blossom', can be used.
    """
    if (circuit_obj is None) == (circuit_path is None):
        raise ValueError('(circuit_obj is None) == (circuit_path is None)')
    if (dem_obj is None) == (dem_path is None):
        raise ValueError('(dem_obj is None) == (dem_path is None)')
    if num_shots == 0:
        return AnonTaskStats()

    with contextlib.ExitStack() as exit_stack:
        start_time = time.monotonic()

        if circuit_path is not None:
            circuit = stim.Circuit.from_file(circuit_path)
        else:
            circuit = circuit_obj
        if tmp_dir is None:
            tmp_dir = exit_stack.enter_context(tempfile.TemporaryDirectory())
        tmp_dir = pathlib.Path(tmp_dir)
        if dem_path is None:
            dem_path = tmp_dir / 'tmp.dem'
            dem_obj.to_file(dem_path)
        dem_path = pathlib.Path(dem_path)

        dets_all_path = tmp_dir / 'sinter_dets.all.b8'
        obs_all_path = tmp_dir / 'sinter_obs.all.b8'
        dets_kept_path = tmp_dir / 'sinter_dets.kept.b8'
        obs_kept_path = tmp_dir / 'sinter_obs.kept.b8'
        predictions_path = tmp_dir / 'sinter_predictions.b8'

        num_dets = circuit.num_detectors
        num_obs = circuit.num_observables

        # Sample data using Stim.
        sampler: stim.CompiledDetectorSampler = circuit.compile_detector_sampler()
        sampler.sample_write(
            num_shots,
            filepath=str(dets_all_path),
            obs_out_filepath=str(obs_all_path),
            format='b8',
            obs_out_format='b8',
        )

        # Postselect, then split into detection event data and observable data.
        if post_mask is None:
            num_det_discards = 0
            dets_used_path = dets_all_path
            obs_used_path = obs_all_path
        else:
            num_det_discards = streaming_post_select(
                num_shots=num_shots,
                num_dets=num_dets,
                num_obs=num_obs,
                dets_in_b8=dets_all_path,
                dets_out_b8=dets_kept_path,
                obs_in_b8=obs_all_path,
                obs_out_b8=obs_kept_path,
                post_mask=post_mask,
                discards_out_b8=None,
            )
            dets_used_path = dets_kept_path
            obs_used_path = obs_kept_path
        num_kept_shots = num_shots - num_det_discards

        # Perform syndrome decoding to predict observables from detection events.
        decoder_obj: Optional[Decoder] = None
        if custom_decoders is not None:
            decoder_obj = custom_decoders.get(decoder)
        if decoder_obj is None:
            decoder_obj = BUILT_IN_DECODERS.get(decoder)
        if decoder_obj is None:
            raise NotImplementedError(f"Unrecognized decoder: {decoder!r}")
        decoder_obj.decode_via_files(
            num_shots=num_kept_shots,
            num_dets=num_dets,
            num_obs=num_obs,
            dem_path=dem_path,
            dets_b8_in_path=dets_used_path,
            obs_predictions_b8_out_path=predictions_path,
            tmp_dir=tmp_dir,
        )

        # Count how many predictions matched the actual observable data.
        num_obs_discards, num_errors = _streaming_count_mistakes(
            num_shots=num_kept_shots,
            num_obs=num_obs,
            obs_in=obs_used_path,
            predictions_in=predictions_path,
            postselected_observable_mask=postselected_observable_mask,
        )

        return AnonTaskStats(
            shots=num_shots,
            errors=num_errors,
            discards=num_obs_discards + num_det_discards,
            seconds=time.monotonic() - start_time,
        )
